#ifndef bqtSymTablesHH
#define bqtSymTablesHH

#include "hash.hh"
#include "autoptr"
#include <map>
#include <set>

typedef autoptr<struct TypeInfo> TypeP;
typedef autoptr<struct UserType> UtypeP;
typedef autoptr<struct Scope> ScopeP;
typedef autoptr<struct SubRoutine> SubFuncP;

extern const TypeP TypeInt, TypeSng, TypeDbl, TypeLng, TypeStr;

struct TypeInfo: public ptrable
{
    enum TypeType
    {
        Integer, Long, Single, Double, String,
        UserType
    } Type;
    
    enum ArrayType
    {
        NotArray,
        StaticArray,      // '$STATIC
        DynamicArray,     // '$DYNAMIC
        ParamArray        // One passed via parameter (number of dimensions unknown)
    };
    
    ArrayType arraytype;     //if array
    unsigned fixlength;      //if non-dynamic string.
    std::string utypename;   //if UserType
    UtypeP members;          //if UserType
    
    // If array:
    std::vector<std::pair<long, unsigned> > dimensions;
    // ^first  = beginning index
    // ^second = dimension (size)
    //   example: 1 TO 5   is (1,5)
    //            0 TO 4   is (0,5)
    
    TypeInfo(TypeType t = Integer)
    : Type(t),arraytype(NotArray),fixlength(0),members(),dimensions()
    { }
    
    bool is_string() const { return Type == String; }
    bool is_num() const    { return Type != String && Type != UserType; }
    bool is_array() const  { return arraytype != NotArray; }
    
    bool operator==(const TypeInfo& b) const;
    bool operator!=(const TypeInfo& b) const { return !operator==(b); }
    
    const std::string Dump() const;
};

#include "expr.hh"

struct UserType: public std::vector<std::pair<std::string, TypeP> >, public ptrable
{
    hash_map<std::string, unsigned> Names;
    
    bool DeclareMember(const std::string& name, TypeP p);
    TypeP GetMember(const std::string& name) const;
    
    bool operator==(const UserType& b)const;
};

struct TypeFlagSet
{
    //union {
        struct{ 
    bool is_subroutine:1;
    bool is_scalar    :1;
    bool is_const     :1;
    bool is_array     :1;
    
    bool is_global_scalar : 1; // imported using SHARED
    bool is_global_array  : 1; // imported using SHARED
    
    // In case of subroutines, they can be DECLARED and later defined.
    bool has_body     :1;
        };
        //unsigned zero : 8;
    //};
    
    // Types and labels not included here; they are in different namespace.
    
    TypeFlagSet() :
        is_subroutine(0), is_scalar(0),
        is_const(0), is_array(0),
        is_global_scalar(0), is_global_array(0),
        has_body(0) {}
};

struct SubParamUsageReport
{
    struct ParamUsage
    {
        bool AssignedTo;
        std::set<std::pair<SubFuncP, unsigned/*param index*/> > PassedAsParam;
        
        ParamUsage(): AssignedTo(false),PassedAsParam() {}
    };
    /* for each param, their status */
    std::vector<ParamUsage> report;
};

/* SUB, FUNCTION, DEF FN */
struct SubRoutine: public ptrable
{
    struct ParamSpec
    {
        std::string name;
        TypeP       type;
        bool        is_array;
        enum { ByRef, ByVal, Seg } passconvention;
        bool        was_modified; // If the param is never modified,
                                  // It can be modified into ByVal.

        ParamSpec() : name(), type(), is_array(false),
          passconvention(ByRef), was_modified(false) { }
        
        void SetByval() { passconvention=ByVal; }
        void SetSeg() { passconvention=Seg; }
    };
    
    TypeP returntype; // 0=SUB, otherwise FUNCTION/DEF
    std::vector<ParamSpec> params;
    bool is_static;
    autoptr<class stmt> code;
    
    int callconvention;     // 0=default, 1=cdecl
    std::string linkername; // ALIAS
    
    ScopeP sub_scope;
    
    SubParamUsageReport params_usage;

public:
    SubRoutine(): returntype(), params(), is_static(false),
                  code(), callconvention(0), linkername(),
                  sub_scope(), params_usage() {}
};

struct LabelDef
{
    bool     defined;
    unsigned dataptr;
};

typedef autoptr<struct VarContextDef> ctxp;
struct VarContextDef: public ptrable
{
    TypeP type;
    bool is_static; // STATIC to the given function
    bool is_shared; // SHARED to the subfunctions (towards children)
    int param_index;
    //stmtp first_use;
    
public:
    VarContextDef(): type(),is_static(false),is_shared(false), param_index(-1)
    //               ,first_use(0)
    {}
};

typedef hash_map<std::string, ctxp> ScopeVarListType;

struct Scope: public ptrable
{
    /*****
      Method of lookup:
        Check constants. Ignore suffix.
          Match? Fail if had wrong suffix. Otherwise, accept.
      
        Check UTYPE.
          Match? Fail if had a suffix. Otherwise, accept.
        Had suffix? Check by suffix.
        Otherwise, check by DEFtype.
    *******/
    
    hash_map<std::string, UtypeP> types;    // Types have no suffix.
    hash_map<std::string, exprp> constants; // Constants have no suffix.
    
    // SUBs, FUNCTIONs, DEFs
    hash_map<std::string, SubFuncP> subroutines;
    
    struct ScopeDef
    {
        ScopeVarListType scalars;
        ScopeVarListType arrays;
    } Utype, Int, Single, Double, Long, String;
    
    hash_map<std::string, TypeFlagSet> usage;
    
    // Labels...
    hash_map<std::string, long double> labelnames;
    std::map<long double, LabelDef> labels;
};

#endif
